// cloudobj.cxx -- routines to generate a "cloud" polygon
// Based on Objects/obj.cxx  
//
// Originally Written by Curtis Olson, started October 1997.
// Adapted for cloud generation by Durk Talsma
//
// Copyright (C) 1997  Curtis L. Olson  - curt@infoplane.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: cloudobj.cxx,v 1.5 2000/02/16 22:01:07 curt Exp $


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef FG_MATH_EXCEPTION_CLASH
#  include <math.h>
#endif

#include <stdio.h>
#include <string.h>

// #if defined ( __sun__ )
// extern "C" void *memmove(void *, const void *, size_t);
// extern "C" void *memset(void *, int, size_t);
// #endif

#include <simgear/compiler.h>

#include STL_STRING
#include <map>			// STL
#include <vector>		// STL
#include <ctype.h>		// isdigit()

#include <simgear/constants.h>
#include <simgear/debug/logstream.hxx>
#include <simgear/math/mat3.h>
#include <simgear/math/fg_geodesy.hxx>
#include <simgear/math/fg_random.h>
#include <simgear/math/point3d.hxx>
#include <simgear/math/polar3d.hxx>
#include <simgear/misc/fgstream.hxx>
#include <simgear/misc/stopwatch.hxx>
#include <simgear/misc/texcoord.hxx>

#include <Main/options.hxx>
#include <Objects/materialmgr.hxx>
#include <Scenery/tileentry.hxx>

#include "cloudobj.hxx"

FG_USING_STD(string);
FG_USING_STD(vector);


typedef vector < int > int_list;
typedef int_list::iterator int_list_iterator;
typedef int_list::const_iterator int_point_list_iterator;


// static double normals[FG_MAX_NODES][3];
// static double tex_coords[FG_MAX_NODES*3][3];

#if 0
// given three points defining a triangle, calculate the normal
static void calc_normal(Point3D p1, Point3D p2, 
			Point3D p3, double normal[3])
{
    double v1[3], v2[3];
    double temp;

    v1[0] = p2[0] - p1[0]; v1[1] = p2[1] - p1[1]; v1[2] = p2[2] - p1[2];
    v2[0] = p3[0] - p1[0]; v2[1] = p3[1] - p1[1]; v2[2] = p3[2] - p1[2];

    MAT3cross_product(normal, v1, v2);
    MAT3_NORMALIZE_VEC(normal,temp);

    // fgPrintf( FG_TERRAIN, FG_DEBUG, "  Normal = %.2f %.2f %.2f\n", 
    //           normal[0], normal[1], normal[2]);
}

#define FG_TEX_CONSTANT 69.0

// Calculate texture coordinates for a given point.
static Point3D calc_tex_coords(const Point3D& node, const Point3D& ref) {
    Point3D cp;
    Point3D pp;
    // double tmplon, tmplat;

    // cout << "-> " << node[0] << " " << node[1] << " " << node[2] << endl;
    // cout << "-> " << ref.x() << " " << ref.y() << " " << ref.z() << endl;

    cp = Point3D( node[0] + ref.x(),
		  node[1] + ref.y(),
		  node[2] + ref.z() );

    pp = fgCartToPolar3d(cp);

    // tmplon = pp.lon() * RAD_TO_DEG;
    // tmplat = pp.lat() * RAD_TO_DEG;
    // cout << tmplon << " " << tmplat << endl;

    pp.setx( fmod(RAD_TO_DEG * FG_TEX_CONSTANT * pp.x(), 11.0) );
    pp.sety( fmod(RAD_TO_DEG * FG_TEX_CONSTANT * pp.y(), 11.0) );

    if ( pp.x() < 0.0 ) {
	pp.setx( pp.x() + 11.0 );
    }

    if ( pp.y() < 0.0 ) {
	pp.sety( pp.y() + 11.0 );
    }

    // cout << pp << endl;

    return(pp);
}
#endif


// Generate a generic ocean tile on the fly
ssgLeaf *fgGenCloudLayer( FGTileEntry *t, double alt ) {
    fgFRAGMENT fragment;
    fragment.init();
    fragment.tile_ptr = t;

    FG_LOG( FG_TERRAIN, FG_DEBUG, "Building Cloud tile");
    ssgSimpleState *state = NULL;
    int i;

    // find Ocean material in the properties list
    if ( ! material_mgr.find( "Cloud", fragment.material_ptr )) {
	FG_LOG( FG_TERRAIN, FG_ALERT, 
		"Ack! unknown usemtl name = " << "Cloud" 
		<< " in fgGenCloudLayer" );
    }

    // set the texture width and height values for this
    // material
    FGMaterial m = fragment.material_ptr->get_m();
    // double tex_width = m.get_xsize();
    // double tex_height = m.get_ysize();

    // set ssgState
    //state = new ssgSimpleState;
    state = fragment.material_ptr->get_state();

    // Calculate center point
    FGBucket b = t->tile_bucket;
    double clon = b.get_center_lon();
    double clat = b.get_center_lat();
    double height = b.get_height();
    double width = b.get_width();

    Point3D center; // = fgGeodToCart(Point3D(clon*DEG_TO_RAD,clat*DEG_TO_RAD,0.0));
    center = t->center;
    //t->center = center;
    fragment.center = center;
    // cout << "center = " << center << endl;;
    
    // Caculate corner vertices
    Point3D geod[4];
    geod[0] = Point3D( clon - width/2.0, clat - height/2.0, alt ); // z should be acuired from the
    geod[1] = Point3D( clon + width/2.0, clat - height/2.0, alt ); // weather database
    geod[2] = Point3D( clon + width/2.0, clat + height/2.0, alt );
    geod[3] = Point3D( clon - width/2.0, clat + height/2.0, alt );

    Point3D rad[4];
    for ( i = 0; i < 4; ++i ) {
	rad[i] = Point3D( geod[i].x() * DEG_TO_RAD, geod[i].y() * DEG_TO_RAD,
			  geod[i].z() );
    }

    Point3D cart[4], rel[4];
    //t->nodes.clear();
    for ( i = 0; i < 4; ++i ) {
    	cart[i] = fgGeodToCart(rad[i]);
	rel[i] = cart[i] - center;
	t->nodes.push_back( rel[i] );
	// cout << "corner " << i << " = " << cart[i] << endl;
    }

    t->ncount += 4;

    // Calculate bounding radius
    //t->bounding_radius = center.distance3D( cart[0] );
    fragment.bounding_radius = t->bounding_radius;
    // cout << "bounding radius = " << t->bounding_radius << endl;

    // Calculate normals
    Point3D normals[4];
    for ( i = 0; i < 4; ++i ) {
	normals[i] = cart[i];
	double length = normals[i].distance3D( Point3D(0.0) );
	normals[i] /= length;
	// cout << "normal = " << normals[i] << endl;
    }

    // Calculate texture coordinates
    point_list geod_nodes;
    geod_nodes.clear();
    for ( i = 0; i < 4; ++i ) {
	geod_nodes.push_back( geod[i] );
    }
    int_list rectangle;
    rectangle.clear();
    for ( i = 0; i < 4; ++i ) {
	rectangle.push_back( i );
    }
    point_list texs = calc_tex_coords( b, geod_nodes, rectangle, 0.125 );

    // Build flight gear structure
    fragment.add_face(0, 1, 2);
    fragment.add_face(0, 2, 3);
    t->fragment_list.push_back(fragment);

    // Build ssg structure
    sgVec3 *vtlist = new sgVec3 [ 4 ];
    t->vec3_ptrs.push_back( vtlist );
    sgVec3 *vnlist = new sgVec3 [ 4 ];
    t->vec3_ptrs.push_back( vnlist );
    sgVec2 *tclist = new sgVec2 [ 4 ];
    t->vec2_ptrs.push_back( tclist );

    for ( i = 0; i < 4; ++i ) {
	sgSetVec3( vtlist[i], 
		   rel[i].x(), rel[i].y(), rel[i].z() );
	sgSetVec3( vnlist[i], 
		   normals[i].x(), normals[i].y(), normals[i].z() );
	sgSetVec2( tclist[i], texs[i].x(), texs[i].y());
    }
    
    unsigned short *vindex = new unsigned short [ 4 ];
    t->index_ptrs.push_back( vindex );
    unsigned short *tindex = new unsigned short [ 4 ];
    t->index_ptrs.push_back( tindex );
    for ( i = 0; i < 4; ++i ) {
	vindex[i] = i;
	tindex[i] = i;
    }

    ssgLeaf *cloud_layer = 
	new ssgVTable ( GL_TRIANGLE_FAN,
			4, vindex, vtlist,
			4, vindex, vnlist,
			4, tindex, tclist,
			0, NULL, NULL ) ;
    //state -> setTranslucent () ;
    //state -> enable ( GL_ALPHA_TEST ) ;
    //state -> setAlphaClamp ( 0.65 ) ;
    state -> enable ( GL_BLEND ) ;
    state->disable(GL_CULL_FACE);
    state -> setTranslucent () ;
    //state -> disable(GL_LIGHTING);
    //state->enable (GL_BLEND);

    cloud_layer->setState( state );

    return cloud_layer;
}
